# This file is placed in the Public Domain.
# pylint: disable=R0903


"commands"


from .object import Object


class Commands:

    "Commands"

    cmds     = Object()
    modnames = Object()

    @staticmethod
    def add(func):
        "add command."
        setattr(Commands.cmds, func.__name__, func)
        if func.__module__ != "__main__":
            setattr(Commands.modnames, func.__name__, func.__module__)


def __dir__():
    return (
        'Commands',
    )



# This file is placed in the Public Domain.
# pylint: disable=R0903,E1102


"logging"


class Logging:

    "Logging"

    filter = []
    out    = None


def debug(txt):
    "print to console."
    for skp in Logging.filter:
        if skp in txt:
            return
    if Logging.out:
        Logging.out(txt)


def __dir__():
    return (
        'Logging',
        'debug'
    )



# This file is placed in the Public Domain.


"utilities"


import datetime
import os
import pathlib
import pwd
import time
import types as rtypes
import _thread


def cdir(pth):
    "create directory."
    path = pathlib.Path(pth)
    path.parent.mkdir(parents=True, exist_ok=True)


def fntime(daystr):
    "convert file name to it's saved time."
    daystr = daystr.replace('_', ':')
    datestr = ' '.join(daystr.split(os.sep)[-2:])
    if '.' in datestr:
        datestr, rest = datestr.rsplit('.', 1)
    else:
        rest = ''
    timed = time.mktime(time.strptime(datestr, '%Y-%m-%d %H:%M:%S'))
    if rest:
        timed += float('.' + rest)
    return timed


def forever():
    "it doesn't stop, until ctrl-c"
    while True:
        try:
            time.sleep(1.0)
        except (KeyboardInterrupt, EOFError):
            _thread.interrupt_main()


def fqn(obj):
    "return full qualified name of an object."
    kin = str(type(obj)).split()[-1][1:-2]
    if kin == "type":
        kin = f"{obj.__module__}.{obj.__name__}"
    return kin


def ident(obj):
    "return an id for an object."
    return os.path.join(fqn(obj), *str(datetime.datetime.now()).split())


def laps(seconds, short=True):
    "show elapsed time."
    txt = ""
    nsec = float(seconds)
    if nsec < 1:
        return f"{nsec:.2f}s"
    yea = 365*24*60*60
    week = 7*24*60*60
    nday = 24*60*60
    hour = 60*60
    minute = 60
    yeas = int(nsec/yea)
    nsec -= yeas*yea
    weeks = int(nsec/week)
    nsec -= weeks*week
    nrdays = int(nsec/nday)
    nsec -= nrdays*nday
    hours = int(nsec/hour)
    nsec -= hours*hour
    minutes = int(nsec/minute)
    nsec -= int(minute*minutes)
    sec = int(nsec)
    if yeas:
        txt += f"{yeas}y"
    if weeks:
        nrdays += weeks * 7
    if nrdays:
        txt += f"{nrdays}d"
    if short and txt:
        return txt.strip()
    if hours:
        txt += f"{hours}h"
    if minutes:
        txt += f"{minutes}m"
    if sec:
        txt += f"{sec}s"
    txt = txt.strip()
    return txt


def modnames(*args):
    "return module names."
    res = []
    for arg in args:
        res.extend([x for x in dir(arg) if not x.startswith("__")])
    return sorted(res)


def named(obj):
    "return a full qualified name of an object/function/module."
    res = None
    if isinstance(obj, rtypes.ModuleType):
        res =  obj.__name__
    typ = type(obj)
    if '__builtins__' in dir(typ):
        res =  obj.__name__
    elif '__self__' in dir(obj):
        res =  f'{obj.__self__.__class__.__name__}.{obj.__name__}'
    elif '__class__' in dir(obj) and '__name__' in dir(obj):
        res =  f'{obj.__class__.__name__}.{obj.__name__}'
    elif '__class__' in dir(obj):
        res =  f"{obj.__class__.__module__}.{obj.__class__.__name__}"
    elif '__name__' in dir(obj):
        res =  f'{obj.__class__.__name__}.{obj.__name__}'
    return res


def pidfile(pid):
    "write the pid to a file."
    if os.path.exists(pid):
        os.unlink(pid)
    path = pathlib.Path(pid)
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(pid, "w", encoding="utf-8") as fds:
        fds.write(str(os.getpid()))


def privileges(username):
    "drop privileges."
    pwnam = pwd.getpwnam(username)
    os.setgid(pwnam.pw_gid)
    os.setuid(pwnam.pw_uid)


def skip(name, skipp):
    "check for skipping"
    for skp in spl(skipp):
        if skp in name:
            return True
    return False


def spl(txt):
    "split comma separated string into a list."
    try:
        res = txt.split(',')
    except (TypeError, ValueError):
        res = txt
    return [x for x in res if x]


def strip(pth, nmr=3):
    "reduce to path with directory."
    return os.sep.join(pth.split(os.sep)[-nmr:])


def __dir__():
    return (
        'cdir',
        'fntime',
        'forever',
        'fqn',
        'ident',
        'laps',
        'modnames',
        'named',
        'pidfile',
        'privileges',
        'skip',
        'spl',
        'strip'
    )



# This file is placed in the Public Domain.
# pylint: disable=C,I,R


"disk"


import os


from .decoder import read
from .encoder import write
from .lock    import disklock
from .workdir import store
from .utils   import ident


def fetch(obj, pth):
    "read object from disk."
    with disklock:
        pth2 = store(pth)
        read(obj, pth2)
        return os.sep.join(pth.split(os.sep)[-3:])


def sync(obj, pth=None):
    "sync object to disk."
    with disklock:
        if pth is None:
            pth = ident(obj)
        pth2 = store(pth)
        write(obj, pth2)
        return pth


def __dir__():
    return (
        'fetch',
        'sync'
    )



# This file is placed in the Public Domain.


"find"


import os


from .default import Default
from .disk    import fetch
from .object  import search, update
from .utils   import fqn, fntime, strip
from .workdir import long, store


def last(obj, selector=None):
    "return last object saved."
    if selector is None:
        selector = {}
    result = sorted(
                    find(fqn(obj), selector),
                    key=lambda x: fntime(x[0])
                   )
    res = None
    if result:
        inp = result[-1]
        update(obj, inp[-1])
        res = inp[0]
    return res


def fns(mtc=""):
    "show list of files."
    dname = ''
    pth = store(mtc)
    for rootdir, dirs, _files in os.walk(pth, topdown=False):
        if dirs:
            for dname in sorted(dirs):
                if dname.count('-') == 2:
                    ddd = os.path.join(rootdir, dname)
                    for fll in os.scandir(ddd):
                        yield strip(os.path.join(ddd, fll))



def find(mtc, selector=None, index=None, deleted=False):
    "find object matching the selector dict."
    clz = long(mtc)
    nrs = -1
    for fnm in sorted(fns(clz), key=fntime):
        obj = Default()
        fetch(obj, fnm)
        if not deleted and '__deleted__' in obj and obj.__deleted__:
            continue
        if selector and not search(obj, selector):
            continue
        nrs += 1
        if index is not None and nrs != int(index):
            continue
        yield (fnm, obj)


def __dir__():
    return (
        'find',
        'fns',
        'last'
    )



# This file is placed in the Public Domain.


"client"


from .command import Commands
from .object  import Object
from .parse   import parse
from .reactor import Reactor


class Client(Reactor):

    "Client"

    cache = Object()
    out = None

    def __init__(self, outer=None):
        Reactor.__init__(self)
        self.register("command", command)
        self.out = outer

    def say(self, _channel, txt):
        "echo on verbose."
        self.raw(txt)

    def raw(self, txt):
        "print to screen."
        if self.out:
            txt = txt.encode('utf-8', 'replace').decode()
            self.out(txt)

    def show(self, evt):
        "show results into a channel."
        for txt in evt.result:
            self.say(evt.channel, txt)


def command(bot, evt):
    "check for and run a command."
    parse(evt)
    func = getattr(Commands.cmds, evt.cmd, None)
    if func:
        func(evt)
        bot.show(evt)
    evt.ready()


def __dir__():
    return (
       'Client',
    )



# This file is placed in the Public Domain.
# pylint: disable=R0902,R0903


"configuration"


import os


from .default import Default


class Config(Default):

    "Config"

    name    = Default.__module__.rsplit(".", maxsplit=2)[-2]
    wdr     = os.path.expanduser(f"~/.{name}")
    pidfile = os.path.join(wdr, f"{name}.pid")

    def __init__(self):
        Default.__init__(self)
        self.name = Config.name
        self.wdr  = Config.wdr
        self.pidfile = Config.pidfile


def __dir__():
    return (
        "Config",
    )



# This file is placed in the Public Domain.
# pylint: disable=C,I,R


"workdir"


import os
import pathlib


from .config import Config
from .utils  import fqn


class Workdir:

    "Workdir"

    fqns = []
    wdr = Config.wdr


def long(name):
    "match from single name to long name."
    split = name.split(".")[-1].lower()
    res = name
    for names in types():
        if split == names.split(".")[-1].lower():
            res = names
            break
    return res


def setwd(path=None):
    Workdir.wdr = path or os.path.expanduser(f"~/.{Config.name}")


def skel():
    "create directory,"
    stor = os.path.join(Workdir.wdr, "store", "")
    path = pathlib.Path(stor)
    path.mkdir(parents=True, exist_ok=True)
    return path


def store(pth=""):
    "return objects directory."
    stor = os.path.join(Workdir.wdr, "store", "")
    if not os.path.exists(stor):
        skel()
    return os.path.join(Workdir.wdr, "store", pth)


def types():
    "return types stored."
    return os.listdir(store())


def whitelist(clz):
    "whitelist classes."
    Workdir.fqns.append(fqn(clz))


def __dir__():
    return (
        'Workdir',
        'setwd',
        'long',
        'skel',
        'store',
        'types',
        'whitelist'
    )



# This file is placed in the Public Domain.
# pylint: disable=W0718


"threads"


import queue
import threading
import time


from .errors import later
from .event  import Event
from .utils  import named


class Thread(threading.Thread):

    "Thread"

    def __init__(self, func, thrname, *args, daemon=True, **kwargs):
        super().__init__(None, self.run, thrname, (), {}, daemon=daemon)
        self._result   = None
        self.name      = thrname or (func and named(func)) or named(self).split(".")[-1]
        self.out       = None
        self.queue     = queue.Queue()
        self.sleep     = None
        self.starttime = time.time()
        if func:
            self.queue.put_nowait((func, args))

    def __iter__(self):
        return self

    def __next__(self):
        yield from dir(self)

    def size(self):
        "return qsize"
        return self.queue.qsize()

    def join(self, timeout=1.0):
        "join this thread."
        super().join(timeout)
        return self._result

    def run(self):
        "run this thread's payload."
        func, args = self.queue.get()
        try:
            self._result = func(*args)
        except Exception as ex:
            later(ex)
            for arg in args:
                if isinstance(arg, Event):
                    arg.ready()


def launch(func, *args, **kwargs):
    "launch a thread."
    nme = kwargs.get("name", named(func))
    thread = Thread(func, nme, *args, **kwargs)
    thread.start()
    return thread


def __dir__():
    return (
        'Thread',
        'launch'
    )



# This file is placed in the Public Domain.


"list of commands"


from ..command import Commands
from ..object  import keys


def cmd(event):
    "list commands."
    event.reply(",".join(sorted(keys(Commands.cmds))))


Commands.add(cmd)



# This file is placed in the Public Domain.


"""OTPCR


SYNOPSIS

    otpcr <cmd> [key=val] [key==val]
    otpcrc
    otpcrd
    otpcrs

DESCRIPTION

    OTPCR holds evidence that king netherlands is doing a
    genocide,  written response where king netherlands confirmed
    taking note of “what i have written”, namely proof that medicine
    he uses in treatement laws like zyprexa, haldol, abilify and
    clozapine are poison that make impotent, is both physical
    (contracted muscles) and mental (make people hallucinate) torture
    and kills members of the victim groups.

    OTPCR contains correspondence with the International Crimina
    Court, asking for arrest of the king of the netherlands, for the
    genocide he is committing with his new treatement laws.

    Current status is a "no basis to proceed" judgement of the
    prosecutor which requires a "basis to prosecute" to have the king
    actually arrested.

INSTALL

    $ pipx install otpcr
    $ pipx ensurepath

USAGE

    without any argument otpcr does nothing

    $ otpcr
    $

    see list of commands

    $ otpcr cmd
    cmd,err,mod,srv,thr,upt

    start a console

    $ otpcrc
    >

    use -i to run init on modules

    $ otpcrc -i
    >

    start daemon

    $ otpcrd
    $

    start as service

    $ otpcrs
    <waits till ctrl-c>

COMMANDS

    cmd - commands
    err - show errors
    mod - available modules
    srv - create service file
    thr - show running threads
    upt - show uptime

SYSTEMD

    $ otpcr srv > otpcr.service
    $ sudo mv otpcr.service /etc/systemd/system/
    $ sudo systemctl enable otpcr --now

FILES

    ~/.otpcr
    ~/.local/bin/otpcr
    ~/.local/bin/otpcrc
    ~/.local/bin/otpcrd
    ~/.local/bin/otpcrs
    ~/.local/pipx/venvs/otpcr/

AUTHOR

    Bart Thate <rssbotd@gmail.com>

COPYRIGHT

    OTPCR is Public Domain.

"""



# This file is placed in the Public Domain.


"list of bots"


rpr = object.__repr__


class Fleet:

    "Fleet"

    bots = []

    @staticmethod
    def all():
        "return all objects."
        return Fleet.bots

    @staticmethod
    def announce(txt):
        "announce on all bots."
        for bot in Fleet.bots:
            if "announce" in dir(bot):
                bot.announce(txt)

    @staticmethod
    def get(orig):
        "return bot."
        res = None
        for bot in Fleet.bots:
            if rpr(bot) == orig:
                res = bot
                break
        return res

    @staticmethod
    def register(obj):
        "add bot."
        Fleet.bots.append(obj)


def __dir__():
    return (
        'Fleet',
    )



# This file is placed in the Public Domain.
# pylint: disable=E1102


"deferred exception handling"


import io
import traceback


class Errors:

    "Errors"

    cache  = []
    errors = []
    out    = None

    @staticmethod
    def format(exc):
        "format an exception"
        res = ""
        stream = io.StringIO(
                             traceback.print_exception(
                                                       type(exc),
                                                       exc,
                                                       exc.__traceback__
                                                      )
                            )
        for line in stream.readlines():
            res += line + "\n"
        return res

    @staticmethod
    def output(exc):
        "check if output function is set."
        if Errors.out:
            text = Errors.format(exc)
            if text not in Errors.cache:
                Errors.cache.append(text)
                Errors.out(text)


def errors():
    "show exceptions"
    for exc in Errors.errors:
        Errors.output(exc)


def later(exc):
    "add an exception"
    excp = exc.with_traceback(exc.__traceback__)
    Errors.errors.append(excp)


def __dir__():
    return (
        'Errors',
        'errors',
        'later'
    )



# This file is placed in the Public Domain.
# pylint: disable=W0401,W0611,W0614,W0622
# ruff: noqa: F401,F403


"interface"


from . import client, command, config, console, decoder, disk, encoder, errors
from . import event, reactor, log, object, parse, repeater, thread, workdir
from . import fleet, timer, utils, find


from .client   import *
from .config   import *
from .command  import *
from .console  import *
from .decoder  import *
from .disk     import *
from .encoder  import *
from .errors   import *
from .event    import *
from .find     import *
from .fleet    import *
from .log      import *
from .main     import *
from .object   import *
from .parse    import *
from .reactor  import *
from .repeater import *
from .thread   import *
from .timer    import *
from .utils    import *
from .workdir  import *


def __dir__():
    return (
        'Broker',
        'CLI',
        'Commands',
        'Config',
        'config',
        'Console',
        'Default',
        'Errors',
        'Event',
        'Handler',
        'Logging',
        'Object',
        'Reactor',
        'Repeater',
        'Thread',
        'Timer',
        'Workdir',
        'boot',
        'broker',
        'banner',
        'command',
        'daemon',
        'debug',
        'errors',
        'event',
        'fetch',
        'find',
        'fns',
        'fntime',
        'getmods',
        'laps',
        'last',
        'later',
        'launch',
        'long',
        'modnames',
        'named',
        'privileges',
        'read',
        'scan',
        'skel',
        'spl',
        'store',
        'strip',
        'sync',
        'wrap',
        'write'
    )



# This module is placed in the Public Domain.


"uptime"


import time


from ..command import Commands
from ..utils   import laps


STARTTIME = time.time()


def upt(event):
    "show uptime."
    event.reply(laps(time.time()-STARTTIME))


Commands.add(upt)



# This file is placed in the Public Domain.
# pylint: disable=R0903


"log text"


import time


from ..command import Commands
from ..disk    import sync
from ..object  import Object
from ..find    import find
from ..utils   import fntime, laps


class Log(Object):

    "Log"

    def __init__(self):
        super().__init__()
        self.txt = ''


def log(event):
    "log text."
    if not event.rest:
        nmr = 0
        for fnm, obj in find('log'):
            lap = laps(time.time() - fntime(fnm))
            event.reply(f'{nmr} {obj.txt} {lap}')
            nmr += 1
        if not nmr:
            event.reply('no log')
        return
    obj = Log()
    obj.txt = event.rest
    sync(obj)
    event.reply('ok')


Commands.add(log)



# This file is placed in the Public Domain.


"locking"


import _thread


disklock   = _thread.allocate_lock()
lock       = _thread.allocate_lock()
jsonlock   = _thread.allocate_lock()


def __dir__():
    return (
        'disklock',
        'lock',
        'jsonlock'
    )



# This file is placed in the Public Domain.


"object decoder"


import json


from .lock   import lock
from .object import Object, construct, update


class ObjectDecoder(json.JSONDecoder):

    "ObjectDecoder"

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, *args, **kwargs)

    def decode(self, s, _w=None):
        "decoding string to object."
        val = json.JSONDecoder.decode(self, s)
        if not val:
            val = {}
        return hook(val)

    def raw_decode(self, s, idx=0):
        "decode partial string to object."
        return json.JSONDecoder.raw_decode(self, s, idx)


def hook(objdict, typ=None):
    "construct object from dict."
    if typ:
        obj = typ()
    else:
        obj = Object()
    construct(obj, objdict)
    return obj


def load(fpt, *args, **kw):
    "load object from file."
    kw["cls"] = ObjectDecoder
    kw["object_hook"] = hook
    return json.load(fpt, *args, **kw)


def loads(string, *args, **kw):
    "load object from string."
    kw["cls"] = ObjectDecoder
    kw["object_hook"] = hook
    return json.loads(string, *args, **kw)


def read(obj, pth):
    "read an object from file path."
    with lock:
        with open(pth, 'r', encoding='utf-8') as ofile:
            update(obj, load(ofile))


def __dir__():
    return (
        'hook',
        'load',
        'loads',
        'read'
    )
